@@ -23,7 +23,7 @@ from qtpy.QtWidgets import (
 import pyqtgraph as pg
 import qtawesome as qta
 
-from ..widgets.settings import get_widget_value
+from ..widgets.settings import get_widget_value, set_widget_value
 from ..stylesheets import QPushButton_style, QScrollArea_style
 from ..widgets import ContainerListWidget, StyledListWidgetItem, ColorPreviewWidget
 
@@ -79,8 +79,11 @@ class PropertyAnalysisDialog(QDialog):
 
     def closeEvent(self, event):
         """Disconnect when dialog closes"""
-        self.cdata.data.vtk_pre_render.disconnect(self._on_render_update)
-        self.cdata.models.vtk_pre_render.disconnect(self._on_render_update)
+        try:
+            self.cdata.data.vtk_pre_render.disconnect(self._on_render_update)
+            self.cdata.models.vtk_pre_render.disconnect(self._on_render_update)
+        except Exception:
+            pass
         super().closeEvent(event)
 
     def _setup_ui(self):
@@ -150,7 +153,8 @@ class PropertyAnalysisDialog(QDialog):
         layout.addWidget(property_group)
 
         options_group = QGroupBox("Visualization Options")
-        options_layout = QVBoxLayout()
+        options_group.setFixedHeight(150)
+        options_layout = QVBoxLayout(options_group)
 
         colormap_layout = QHBoxLayout()
         colormap_layout.addWidget(QLabel("Color Map:"))
@@ -172,7 +176,6 @@ class PropertyAnalysisDialog(QDialog):
         options_layout.addLayout(colormap_layout)
         options_layout.addWidget(self.color_preview)
         options_layout.addLayout(checkbox_layout)
-        options_group.setLayout(options_layout)
         layout.addWidget(options_group)
 
         # Dialog Control Buttons
@@ -242,6 +245,10 @@ class PropertyAnalysisDialog(QDialog):
         self.plot_widget.setBackground(None)
         plot_layout.addWidget(self.plot_widget)
 
+        options_group = QGroupBox("Visualization Options")
+        options_group.setFixedHeight(150)
+        options_layout = QVBoxLayout(options_group)
+
         strat_layout = QHBoxLayout()
         self.plot_title = QLabel("Stratification")
         self.plot_mode_combo = QComboBox()
@@ -249,7 +256,7 @@ class PropertyAnalysisDialog(QDialog):
         self.plot_mode_combo.currentTextChanged.connect(self._update_plot)
         strat_layout.addWidget(self.plot_title)
         strat_layout.addWidget(self.plot_mode_combo)
-        plot_layout.addLayout(strat_layout)
+        options_layout.addLayout(strat_layout)
 
         alpha_layout = QHBoxLayout()
         alpha_layout.addWidget(QLabel("Alpha:"))
@@ -258,7 +265,7 @@ class PropertyAnalysisDialog(QDialog):
         self.alpha_slider.setValue(128)
         self.alpha_slider.valueChanged.connect(self._update_plot)
         alpha_layout.addWidget(self.alpha_slider)
-        plot_layout.addLayout(alpha_layout)
+        options_layout.addLayout(alpha_layout)
 
         colormap_layout = QHBoxLayout()
         colormap_layout.addWidget(QLabel("Color Palette:"))
@@ -267,8 +274,10 @@ class PropertyAnalysisDialog(QDialog):
         self.analysis_colormap_combo.addItems(self.color_preview.colormaps)
         self.analysis_colormap_combo.currentTextChanged.connect(self._update_plot)
         colormap_layout.addWidget(self.analysis_colormap_combo)
-        plot_layout.addLayout(colormap_layout)
+        options_layout.addLayout(colormap_layout)
+
         layout.addWidget(plot_group)
+        layout.addWidget(options_group)
 
         # Dialog Control Buttons
         button_layout = QHBoxLayout()
@@ -345,6 +354,7 @@ class PropertyAnalysisDialog(QDialog):
                 "To Camera",
                 "To Cluster",
                 "To Model",
+                "To Self",
             ],
             "Surface": [
                 "Curvature",
@@ -374,6 +384,7 @@ class PropertyAnalysisDialog(QDialog):
             "To Camera": "distance",
             "To Cluster": "distance",
             "To Model": "distance",
+            "To Self": "distance",
             # Surface
             "Curvature": "mesh_curvature",
             "Edge Length": "mesh_edge_length",
@@ -397,6 +408,7 @@ class PropertyAnalysisDialog(QDialog):
         }
         previous_text = self.property_combo.currentText()
 
+        self.property_combo.blockSignals(True)
         self.property_combo.clear()
         self.property_combo.addItems(properties.get(category, []))
         if previous_text is not None:
@@ -406,11 +418,20 @@ class PropertyAnalysisDialog(QDialog):
 
         if self.property_combo.count() > 0:
             self._update_options(self.property_combo.currentText())
+        self.property_combo.blockSignals(False)
 
     def _update_options(self, property_name: str = None):
         if property_name is None:
             property_name = self.property_combo.currentText()
 
+        previous_parameters = {}
+        if hasattr(self, "option_widgets"):
+            previous_parameters = {
+                k: get_widget_value(w)
+                for k, w in self.option_widgets.items()
+                if not isinstance(w, (QListWidget, ContainerListWidget))
+            }
+
         while self.property_options_layout.rowCount() > 0:
             self.property_options_layout.removeRow(0)
 
@@ -512,21 +533,36 @@ class PropertyAnalysisDialog(QDialog):
             self.option_widgets["k_start"] = neighbor_start
             self.option_widgets["k"] = neighbor_end
 
-        elif property_name == "To Cluster":
-            target_group = QGroupBox("Target Clusters")
+        elif property_name in ("To Cluster", "To Self"):
+
+            target_group = QGroupBox("Options")
             target_layout = QVBoxLayout(target_group)
 
-            target_list = _populate_list(self.cdata.format_datalist("data"))
-            target_layout.addWidget(target_list)
+            if property_name == "To Cluster":
 
-            # Checkboxes
-            all_targets_checkbox = QCheckBox("Compare to All")
-            include_self_checkbox = QCheckBox("Include Within-Cluster Distance")
+                target_list = _populate_list(self.cdata.format_datalist("data"))
+                target_layout.addWidget(target_list)
 
-            checkbox_layout = QHBoxLayout()
-            checkbox_layout.addWidget(all_targets_checkbox)
-            checkbox_layout.addWidget(include_self_checkbox)
-            target_layout.addLayout(checkbox_layout)
+                # Checkboxes
+                all_targets_checkbox = QCheckBox("Compare to All")
+                include_self_checkbox = QCheckBox("Include Within-Cluster Distance")
+                all_targets_checkbox.stateChanged.connect(
+                    lambda state: self.toggle_all_targets(state, target_list)
+                )
+
+                checkbox_layout = QHBoxLayout()
+                checkbox_layout.addWidget(all_targets_checkbox)
+                checkbox_layout.addWidget(include_self_checkbox)
+                target_layout.addLayout(checkbox_layout)
+
+                self.option_widgets["queries"] = target_list
+                self.option_widgets["include_self"] = include_self_checkbox
+                self.option_widgets["compare_to_all"] = all_targets_checkbox
+
+            if property_name == "To Self":
+                self_checkbox = QCheckBox()
+                self_checkbox.setChecked(True)
+                self.option_widgets["only_self"] = self_checkbox
 
             # KNN range
             neighbor_layout = QHBoxLayout()
@@ -544,6 +580,8 @@ class PropertyAnalysisDialog(QDialog):
             neighbor_end.setRange(1, 255)
             neighbor_end.setValue(1)
 
+            neighbor_start.valueChanged.connect(lambda x: neighbor_end.setRange(x, 255))
+
             knn_layout.addWidget(neighbor_start)
             knn_layout.addWidget(neighbor_to_label)
             knn_layout.addWidget(neighbor_end)
@@ -551,14 +589,8 @@ class PropertyAnalysisDialog(QDialog):
             neighbor_layout.addLayout(knn_layout)
             target_layout.addLayout(neighbor_layout)
 
-            all_targets_checkbox.stateChanged.connect(
-                lambda state: self.toggle_all_targets(state, target_list)
-            )
-
             self.property_options_layout.addRow(target_group)
 
-            self.option_widgets["queries"] = target_list
-            self.option_widgets["include_self"] = include_self_checkbox
             self.option_widgets["k_start"] = neighbor_start
             self.option_widgets["k"] = neighbor_end
 
@@ -580,6 +612,13 @@ class PropertyAnalysisDialog(QDialog):
 
             self.property_options_layout.addRow(target_group)
             self.option_widgets["queries"] = target_list
+            self.option_widgets["compare_to_all"] = all_targets_checkbox
+
+        # Avoid re-entering parameters over and over
+        for k in self.option_widgets.keys():
+            if k not in previous_parameters:
+                continue
+            set_widget_value(self.option_widgets[k], previous_parameters[k])
 
     def toggle_all_targets(self, state, target_list):
         target_list.setEnabled(not bool(state))
@@ -820,12 +859,18 @@ class PropertyAnalysisDialog(QDialog):
         if not data_series:
             return None
 
-        self.plot_widget.clear()
-        all_scalar = not isinstance(all_values[0], np.ndarray)
-        if all_scalar:
-            all_values = np.asarray(all_values)
-            return self._create_categorical_plot(data_series, all_values, plot_type)
-        return self._create_plot(data_series, all_values, plot_mode, plot_type)
+        try:
+            self.plot_widget.setUpdatesEnabled(False)
+
+            self.plot_widget.clear()
+            all_scalar = not isinstance(all_values[0], np.ndarray)
+            if all_scalar:
+                all_values = np.asarray(all_values)
+                self._create_categorical_plot(data_series, all_values, plot_type)
+            else:
+                self._create_plot(data_series, all_values, plot_mode, plot_type)
+        finally:
+            self.plot_widget.setUpdatesEnabled(True)
 
     def _create_categorical_plot(self, data_series, values, plot_type):
         """Create a categorical plot with names on x-axis for single values"""
@@ -865,9 +910,7 @@ class PropertyAnalysisDialog(QDialog):
                     name=name,
                 )
             plot.addItem(scatter)
-
-        legend = plot.addLegend(offset=(-10, 10))
-        legend.setPos(plot.getViewBox().screenGeometry().width() - 20, 0)
+        plot.addLegend(offset=(-10, 10))
 
     def _create_plot(self, data_series, all_values, plot_mode, plot_type):
         """Create either histogram, density or line plot based on plot_type"""
@@ -900,8 +943,8 @@ class PropertyAnalysisDialog(QDialog):
             plot.setLabel("left", y_label)
             plot.setLabel("bottom", x_label)
 
-            legend = plot.addLegend(offset=(-10, 10))
-            legend.setPos(plot.getViewBox().screenGeometry().width() - 20, 0)
+            plot.disableAutoRange()
+            plot.addLegend(offset=(-10, 10))
 
             for i, (name, obj, values, color) in enumerate(data_series):
                 if plot_type == "Histogram":
@@ -951,6 +994,9 @@ class PropertyAnalysisDialog(QDialog):
                     )
 
                 plot.addItem(item)
+
+            plot.enableAutoRange()
+            plot.autoRange()
             return None
 
         # For separate plots mode
@@ -1036,6 +1082,8 @@ class PropertyAnalysisDialog(QDialog):
 
     def _export_plot(self):
         """Save the current plot as an image"""
+        from pyqtgraph.exporters import ImageExporter
+
         file_path, _ = QFileDialog.getSaveFileName(
             self, "Save Plot", "", "PNG Files (*.png);;All Files (*.*)"
         )
@@ -1044,7 +1092,7 @@ class PropertyAnalysisDialog(QDialog):
             return None
 
         try:
-            exporter = pg.exporters.ImageExporter(self.plot_widget.scene())
+            exporter = ImageExporter(self.plot_widget.scene())
             exporter.parameters()["width"] = 1920
             exporter.parameters()["height"] = 1080
             exporter.parameters()["antialias"] = True
