@@ -7,6 +7,7 @@ Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
 """
 
 import warnings
+from uuid import uuid4
 from typing import Tuple, List, Dict
 
 import vtk
@@ -57,6 +58,8 @@ class Geometry:
         vertex_properties=None,
         **kwargs,
     ):
+        self.uuid = str(uuid4())
+
         self._points = vtk.vtkPoints()
         self._points.SetDataTypeToFloat()
 
@@ -149,6 +152,7 @@ class Geometry:
             "appearance": self._appearance,
             "representation": self._representation,
             "vertex_properties": self.vertex_properties,
+            "uuid": self.uuid,
         }
 
     def __setstate__(self, state):
@@ -160,19 +164,22 @@ class Geometry:
         state : dict
             State dictionary to restore from.
         """
+        uuid = state.pop("uuid", None)
         visible = state.pop("visible", True)
         appearance = state.pop("appearance", {})
         self.__init__(**state)
         self.set_visibility(visible)
 
-        representation = state.get("representation", False)
-        if representation:
-            self.change_representation(representation)
+        if uuid is not None:
+            self.uuid = uuid
 
+        # Required to support loading VolumeGeometries
+        if state.get("representation") != self._representation:
+            self.change_representation(state.get("representation"))
         self.set_appearance(**appearance)
 
     def __getitem__(self, idx):
-        """Array-like indexing using int/bool numpy arrays, slices or ellipses."""
+        """Array-like indexing using int/bool numpy arrays, slices or ellipsis."""
         if isinstance(idx, (int, np.integer)):
             idx = [idx]
         elif isinstance(idx, slice) or idx is ...:
@@ -181,37 +188,23 @@ class Geometry:
         idx = np.asarray(idx)
         if idx.dtype == bool:
             idx = np.where(idx)[0]
+        idx = idx[idx < self.get_number_of_points()]
 
-        normals = None
-        if isinstance(self.normals, np.ndarray):
-            normals = self.normals.copy()
-            if np.max(idx) < normals.shape[0]:
-                normals = normals[idx].copy()
+        state = self.__getstate__()
+        data_array = ("points", "normals", "quaternions")
+        for key in data_array:
+            if (value := state.get(key)) is not None:
+                state[key] = np.asarray(value)[idx].copy()
 
-        quaternions = self._data.GetPointData().GetArray("OrientationQuaternion")
-        if quaternions is not None:
-            quaternions = self.quaternions.copy()
-            if np.max(idx) < quaternions.shape[0]:
-                quaternions = quaternions[idx].copy()
+        if (vertex_properties := state.get("vertex_properties")) is not None:
+            state["vertex_properties"] = vertex_properties[idx]
 
-        if (vertex_properties := self.vertex_properties) is not None:
-            vertex_properties = vertex_properties[idx]
+        if "meta" in state:
+            state["meta"] = state["meta"].copy()
 
-        ret = Geometry(
-            points=self.points[idx].copy(),
-            normals=normals,
-            quaternions=quaternions,
-            color=self._appearance["base_color"],
-            sampling_rate=self._sampling_rate,
-            meta=self._meta.copy(),
-            vertex_properties=vertex_properties,
-        )
-        ret.set_visibility(self.visible)
-
-        # Avoid clashes from properties of classes inheriting from Geometry
-        ret._appearance.update(
-            {k: v for k, v in self._appearance.items() if k in ret._appearance}
-        )
+        _ = state.pop("uuid", None)
+        ret = self.__class__.__new__(self.__class__)
+        ret.__setstate__(state)
         return ret
 
     @classmethod
@@ -237,8 +230,11 @@ class Geometry:
         if not len(geometries):
             raise ValueError("No geometries provided for merging")
 
-        points, quaternions, normals = [], [], []
+        geometries = [x for x in geometries if isinstance(x, Geometry)]
+        if len(geometries) == 1:
+            return geometries[0]
 
+        points, quaternions, normals = [], [], []
         has_quat = any(
             x._data.GetPointData().GetArray("OrientationQuaternion") is not None
             for x in geometries
@@ -270,7 +266,7 @@ class Geometry:
             meta=geometries[0]._meta.copy(),
         )
         ret.set_visibility(any(x.visible for x in geometries))
-        ret._appearance.update(geometries[0]._appearance)
+        ret.set_appearance(**geometries[0]._appearance)
         return ret
 
     @property
@@ -442,7 +438,9 @@ class Geometry:
         """
         if color is None:
             color = self._appearance["base_color"]
-        self.color_points(range(self._points.GetNumberOfPoints()), color=color)
+        self.color_points(
+            np.arange(self._points.GetNumberOfPoints(), dtype=np.int32), color=color
+        )
 
     def set_visibility(self, visibility: bool = True):
         """
@@ -577,9 +575,11 @@ class Geometry:
         use_point : bool, optional
             Whether to use point data for scalar mode, by default False.
         """
-        scalars = np.asarray(scalars).reshape(-1)
+        scalars = np.asarray(scalars).ravel()
         if scalars.size == 1:
-            scalars = np.repeat(scalars, self.points.shape[0])
+            scalars = np.full(
+                (self.get_number_of_points()), fill_value=scalars, dtype=scalars.dtype
+            )
 
         if scalars.size != self.points.shape[0]:
             return None
@@ -663,14 +663,14 @@ class Geometry:
 
         Parameters
         ----------
-        point_ids : set
+        point_ids : np.ndarray
             Set of point indices to color
         color : tuple of float
             RGB color values (0-1) to apply to selected points
         """
         n_points = self._points.GetNumberOfPoints()
         if not isinstance(point_ids, np.ndarray):
-            point_ids = np.array(list(point_ids), dtype=np.int32)
+            point_ids = np.asarray(point_ids, dtype=np.int32)
 
         point_ids = point_ids.astype(np.int32, copy=False)
         point_ids = point_ids[point_ids < n_points]
@@ -995,6 +995,13 @@ class VolumeGeometry(Geometry):
         mapper.OrientOn()
         self._actor.SetMapper(mapper)
 
+        if "upper_quantile" in kwargs and "lower_quantile" in kwargs:
+            self.update_isovalue_quantile(
+                upper_quantile=kwargs.get("upper_quantile"),
+                lower_quantile=kwargs.get("lower_quantile"),
+            )
+        self._representation = "volume"
+
     def __getstate__(self):
         state = super().__getstate__()
 
@@ -1052,13 +1059,10 @@ class VolumeGeometry(Geometry):
         upper_value = np.quantile(self._raw_volume, self._upper_quantile)
         return self.update_isovalue(upper=upper_value, lower=lower_value)
 
-    def change_representation(self, *args, **kwargs) -> int:
-        return -1
-
-    def set_appearance(self, isovalue_percentile=0.99, **kwargs):
+    def set_appearance(self, isovalue_percentile=99.5, **kwargs):
         if hasattr(self, "_raw_volume"):
             self._appearance["isovalue_percentile"] = isovalue_percentile
-            self.update_isovalue_quantile(upper_quantile=isovalue_percentile)
+            self.update_isovalue_quantile(upper_quantile=isovalue_percentile / 100)
         super().set_appearance(**kwargs)
 
 
