@@ -8,11 +8,11 @@ Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
 """
 import os
 from typing import List
-from os.path import splitext, basename
+from os.path import extsep, basename
 
 import vtk
 import numpy as np
-from qtpy.QtCore import Qt, QEvent, QSize
+from qtpy.QtCore import Qt, QEvent, QSize, QTimer
 from qtpy.QtWidgets import (
     QApplication,
     QMainWindow,
@@ -34,6 +34,9 @@ from qtpy.QtGui import (
     QGuiApplication,
     QActionGroup,
     QKeyEvent,
+    QDropEvent,
+    QCursor,
+    QDragEnterEvent,
 )
 import qtawesome as qta
 from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
@@ -191,6 +194,72 @@ class App(QMainWindow):
         self.escape_shortcut = QShortcut(Qt.Key.Key_Escape, self.vtk_widget)
         self.escape_shortcut.activated.connect(self.handle_escape_key)
 
+        QTimer.singleShot(2000, self._check_for_updates)
+
+        self.setAcceptDrops(True)
+        self._drag_active = False
+
+    def dragEnterEvent(self, event: QDragEnterEvent):
+        if event.mimeData().hasUrls():
+            valid_files = False
+            for url in event.mimeData().urls():
+                if url.isLocalFile():
+                    valid_files = True
+                    break
+
+            if valid_files:
+                event.acceptProposedAction()
+                self._drag_active = True
+                QApplication.setOverrideCursor(QCursor(Qt.DragCopyCursor))
+                self.setStyleSheet(
+                    self.styleSheet()
+                    + """
+                    QMainWindow {
+                        border: 2px dashed rgba(99, 102, 241, 0.3);
+                    }
+                    """
+                )
+        return super().dragEnterEvent(event)
+
+    def dragLeaveEvent(self, event):
+        self._drag_active = False
+        QApplication.restoreOverrideCursor()
+        self._update_style()
+
+    def dropEvent(self, event: QDropEvent):
+        self._drag_active = False
+        QApplication.restoreOverrideCursor()
+        self._update_style()
+
+        if not event.mimeData().hasUrls():
+            event.ignore()
+            return None
+
+        event.acceptProposedAction()
+
+        file_paths = []
+        for url in event.mimeData().urls():
+            if url.isLocalFile():
+                file_paths.append(url.toLocalFile())
+
+        if not file_paths:
+            return None
+
+        session_files = [f for f in file_paths if f.lower().endswith(".pickle")]
+        data_files = [f for f in file_paths if not f.lower().endswith(".pickle")]
+
+        if session_files:
+            if len(session_files) > 1:
+                QMessageBox.warning(
+                    self,
+                    "Multiple Session Files",
+                    "Only one session file can be loaded at a time. ",
+                )
+            self._load_session(session_files[0])
+
+        if len(data_files):
+            self._open_files(data_files)
+
     def sizeHint(self):
         """Provide the preferred size for the main window."""
         screen = QGuiApplication.primaryScreen().geometry()
@@ -431,6 +500,12 @@ class App(QMainWindow):
 
         self.volume_dock = None
         self.volume_viewer = MultiVolumeViewer(self.vtk_widget, legend=self.legend)
+        self.cdata.data.render_update.connect(
+            self.volume_viewer.primary.handle_projection_change
+        )
+        self.cdata.models.render_update.connect(
+            self.volume_viewer.primary.handle_projection_change
+        )
 
         self.cursor_handler = CursorModeHandler(self.vtk_widget)
         self.axes_widget = AxesWidget(self.renderer, self.interactor)
@@ -719,9 +794,15 @@ class App(QMainWindow):
         xz_action = QAction("XZ-Plane", self)
         xz_action.setText("Front View (XZ)\tc")
         xz_action.triggered.connect(lambda: self.simulate_key_press("c"))
+
+        flip_action = QAction("Flip View lambda", self)
+        flip_action.setText("Flip View Axis \tv")
+        flip_action.triggered.connect(lambda: self.simulate_key_press("v"))
+
         view_menu.addAction(xy_action)
         view_menu.addAction(yz_action)
         view_menu.addAction(xz_action)
+        view_menu.addAction(flip_action)
         view_menu.addSeparator()
 
         view_menu.addAction(self.volume_action)
@@ -1012,9 +1093,7 @@ class App(QMainWindow):
         self.set_camera_view("z")
 
     def _open_files(self, filenames: List[str]):
-        from mosaic.formats import open_file
-        from mosaic.meshing import to_open3d
-        from mosaic.parametrization import TriangularMesh
+        from .formats import open_file
 
         if isinstance(filenames, str):
             filenames = [
@@ -1044,9 +1123,8 @@ class App(QMainWindow):
                         raise ValueError("Use Load Session to open session files.")
                     raise e
 
-                base, _ = splitext(basename(filename))
+                base, _ = basename(filename).split(extsep, 1)
                 use_index = len(container) > 1
-
                 if len(container) > 1000:
                     reply = QMessageBox.question(
                         self,
@@ -1074,7 +1152,7 @@ class App(QMainWindow):
                         if not show_large_file_warning():
                             continue
 
-                    name = base if not use_index else f"{index}_base"
+                    name = base if not use_index else f"{index}_{base}"
                     if data.faces is None:
                         index = self.cdata.data.add(
                             points=data.vertices,
@@ -1085,6 +1163,9 @@ class App(QMainWindow):
                         )
                         self.cdata._data.data[index]._meta["name"] = name
                     else:
+                        from .meshing import to_open3d
+                        from .parametrization import TriangularMesh
+
                         index = self.cdata._add_fit(
                             fit=TriangularMesh(to_open3d(data.vertices, data.faces)),
                             sampling_rate=sampling,
@@ -1166,6 +1247,23 @@ class App(QMainWindow):
             return self._load_session(file_path)
         return self._open_files([file_path])
 
+    def _check_for_updates(self):
+        from .dialogs import UpdateChecker, UpdateDialog
+        from .__version__ import __version__
+
+        def _show_update_dialog(latest_version, release_notes):
+            if Settings.ui.skipped_version == latest_version:
+                return None
+            dialog = UpdateDialog(
+                __version__, latest_version, release_notes, parent=self
+            )
+            dialog.exec()
+
+        # We assign the thread to keep it alive
+        self.update_checker = UpdateChecker(__version__)
+        self.update_checker.update_available.connect(_show_update_dialog)
+        self.update_checker.start()
+
 
 def show_large_file_warning() -> bool:
     if Settings.warnings.suppress_large_file_warning:
