@@ -24,13 +24,12 @@ from qtpy.QtWidgets import (
 from qtpy.QtCore import (
     Qt,
     QObject,
-    QItemSelection,
-    QItemSelectionModel,
     Signal,
     QEvent,
 )
 from .parallel import submit_task
 
+from .widgets.container_list import StyledTreeWidgetItem
 
 __all__ = ["DataContainerInteractor"]
 
@@ -56,10 +55,10 @@ class DataContainerInteractor(QObject):
         self.data_list.itemChanged.connect(self._on_item_renamed)
         self.data_list.itemSelectionChanged.connect(self._on_cluster_selection_changed)
 
-        self.data_list.list_widget.setContextMenuPolicy(
+        self.data_list.tree_widget.setContextMenuPolicy(
             Qt.ContextMenuPolicy.CustomContextMenu
         )
-        self.data_list.list_widget.customContextMenuRequested.connect(
+        self.data_list.tree_widget.customContextMenuRequested.connect(
             self._show_context_menu
         )
 
@@ -70,8 +69,17 @@ class DataContainerInteractor(QObject):
 
         self.set_coloring_mode("default")
 
-    def _get_selected_indices(self):
-        return sorted([item.row() for item in self.data_list.selectedIndexes()])
+    def _get_selected_uuids(self):
+        """Get UUIDs of selected items."""
+        uuids = []
+        for item in self.data_list.selected_items():
+            if uuid := item.metadata.get("uuid"):
+                uuids.append(uuid)
+        return uuids
+
+    def get_selected_geometries(self):
+        ret = [self.container.get(uuid) for uuid in self._get_selected_uuids()]
+        return [x for x in ret if x is not None]
 
     def attach_area_picker(self):
         self.interactor = self.vtk_widget.GetRenderWindow().GetInteractor()
@@ -132,11 +140,15 @@ class DataContainerInteractor(QObject):
         return super().eventFilter(watched_obj, event)
 
     def _on_item_renamed(self, item):
-        index = self.data_list.row(item)
-        if self.container._index_ok(index):
-            self.container.data[index]._meta["name"] = item.text()
-        self.data_changed.emit()
-        self.render()
+        if (uuid := item.metadata.get("uuid")) is None:
+            return None
+
+        if (geometry := self.container.get(uuid)) is not None:
+            # Consider adding bidrecitional uuid lookup to render instead
+            if item.text() != geometry._meta.get("name"):
+                geometry._meta["name"] = item.text()
+                self.data_changed.emit()
+                self.render()
 
     def next_color(self):
         if not hasattr(self, "colors"):
@@ -146,9 +158,6 @@ class DataContainerInteractor(QObject):
         self.colors.append(color)
         return color
 
-    def get_geometry(self, index):
-        return self.container.get(index)
-
     def set_coloring_mode(self, mode: str):
         from .utils import get_cmap
 
@@ -161,14 +170,20 @@ class DataContainerInteractor(QObject):
             self.colors = [cmap(i)[:3] for i in range(cmap.N)]
 
         for i in range(len(self.container)):
-            self.container.update_appearance([i], {"base_color": self.next_color()})
+            if (geometry := self.container.get(i)) is None:
+                continue
+            self.container.update_appearance(
+                [i], geometry._appearance | {"base_color": self.next_color()}
+            )
         self.container.highlight([])
         return self.render_vtk()
 
     def add(self, *args, **kwargs):
         if kwargs.get("color", None) is None:
             kwargs["color"] = self.next_color()
-        return self.container.add(*args, **kwargs)
+        ret = self.container.add(*args, **kwargs)
+        self.data_changed.emit()
+        return ret
 
     def add_selection(self, selected_point_ids: Dict[vtk.vtkActor, np.ndarray]) -> int:
         """Add new cloud from selected points.
@@ -186,31 +201,29 @@ class DataContainerInteractor(QObject):
         from .geometry import Geometry
 
         new_cluster, remove_cluster = [], []
-        for index, point_ids in selected_point_ids.items():
-            if (geometry := self.get_geometry(index)) is None:
+        for uuid, point_ids in selected_point_ids.items():
+            if (geometry := self.container.get(uuid)) is None:
                 continue
 
-            n_points = geometry.points.shape[0]
+            n_points = geometry.get_number_of_points()
             if not geometry.visible or n_points == 0 or point_ids.size == 0:
                 continue
 
             new_cluster.append(geometry[point_ids])
-
-            inverse = np.setdiff1d(np.arange(n_points), point_ids)
+            inverse = np.setdiff1d(np.arange(n_points), point_ids, assume_unique=True)
             if inverse.size != 0:
-                self.container.data[index] = geometry[inverse]
+                self.container.update(uuid, geometry[inverse])
             else:
                 # All points were selected, mark for removal
-                remove_cluster.append(index)
+                remove_cluster.append(geometry)
 
         self.container.remove(remove_cluster)
-
         if len(new_cluster):
             return self.add(Geometry.merge(new_cluster))
         return -1
 
     def _add_point(self, point):
-        if (geometry := self.get_geometry(self._active_cluster)) is None:
+        if (geometry := self.container.get(self._active_cluster)) is None:
             return -1
 
         # We call swap data to automatically handle other Geometry attributes
@@ -226,33 +239,45 @@ class DataContainerInteractor(QObject):
         self._active_cluster = None
         self._interaction_mode = "draw"
 
-        active_clusters = list(set(self._get_selected_indices()))
-        if len(active_clusters) > 1:
-            print("Can only add points if a single cluster is selected.")
-            return -1
-        elif len(active_clusters) == 0:
-            new_cluster = self.add(points=np.empty((0, 3), dtype=np.float32))
-            active_clusters = [new_cluster]
-
-        self._active_cluster = active_clusters[0]
+        new_cluster_index = self.add(points=np.empty((0, 3), dtype=np.float32))
+        self._active_cluster = self.container.get(new_cluster_index).uuid
 
     def activate_picking_mode(self):
         self._interaction_mode = "pick"
 
-    def set_selection(self, selected_indices):
-        selection = QItemSelection()
-        for index in set(selected_indices):
-            index = self.data_list.model().index(index, 0)
-            selection.select(index, index)
+    def set_selection_by_uuid(self, uuids):
+        """
+        Set selection by UUIDs.
 
-        selection_model_flag = QItemSelectionModel.SelectionFlag
-        self.data_list.selectionModel().select(
-            selection, selection_model_flag.Clear | selection_model_flag.Select
-        )
-        self._on_cluster_selection_changed()
+        Parameters
+        ----------
+        uuids : list of str
+            UUIDs to select
+        """
+        items = []
+        for item, parent, _ in self.data_list.traverse(reverse=False):
+            if item.metadata.get("uuid") in uuids:
+                items.append(item)
+
+        self.data_list._set_selection(items)
+        self._highlight_selection()
 
     def _on_cluster_selection_changed(self):
-        self.container.highlight(self._get_selected_indices())
+        # This is of course not ideal but prevents unhighlight/highlight
+        # when clicking on groups due to itemClicked behaviour. So we
+        # handle actual deselect all using _highlight_selection
+        if not len(self._get_selected_uuids()):
+            return None
+        self._highlight_selection()
+
+    def deselect(self):
+        """Deselect on right-click"""
+        self.data_list.clearSelection()
+        self._highlight_selection()
+        self.deselect_points()
+
+    def _highlight_selection(self):
+        self.container.highlight(self._get_selected_uuids())
         self.render_vtk()
 
     def _on_area_pick(self, obj, event):
@@ -262,25 +287,26 @@ class DataContainerInteractor(QObject):
 
         interactor = self.vtk_widget.GetRenderWindow().GetInteractor()
         if not interactor.GetShiftKey():
-            self.point_selection.clear()
+            self.deselect_points()
 
-        for i, cluster in enumerate(self.container.data):
-            extractor.SetInputData(cluster._data)
+        for geometry in self.container.data:
+            extractor.SetInputData(geometry._data)
             extractor.Update()
 
             output = extractor.GetOutput()
             if output.GetNumberOfPoints() == 0:
                 continue
 
+            uuid = geometry.uuid
             selected_ids = output.GetPointData().GetArray("vtkOriginalPointIds")
             if selected_ids and selected_ids.GetNumberOfTuples() > 0:
                 ids = vtk.util.numpy_support.vtk_to_numpy(selected_ids)
 
-                if i not in self.point_selection:
-                    self.point_selection[i] = np.array([], dtype=np.int32)
+                if uuid not in self.point_selection:
+                    self.point_selection[uuid] = np.array([], dtype=np.int32)
 
-                union = np.union1d(ids, self.point_selection[i])
-                self.point_selection[i] = union.astype(np.int32, copy=False)
+                union = np.union1d(ids, self.point_selection[uuid])
+                self.point_selection[uuid] = union.astype(np.int32, copy=False)
 
         self.highlight_selected_points(color=None)
 
@@ -293,8 +319,8 @@ class DataContainerInteractor(QObject):
         actors = self.container.get_actors()
         if picked_actor in actors:
             index = actors.index(picked_actor)
-            selected_indices = self._get_selected_indices()
-            self.set_selection([index, *selected_indices])
+            uuid = self.container.get(index).uuid
+            self.set_selection_by_uuid([uuid, *self._get_selected_uuids()])
         return None
 
     def _show_context_menu(self, position):
@@ -302,6 +328,8 @@ class DataContainerInteractor(QObject):
         if not item:
             return -1
 
+        # Make sure right click also selects group members
+        self.data_list._select_group_children(item)
         context_menu = QMenu(self.data_list)
 
         show_action = QAction("Show", self.data_list)
@@ -330,9 +358,7 @@ class DataContainerInteractor(QObject):
             "Surface",
             "Wireframe",
         ]
-        for index in self._get_selected_indices():
-            if (geometry := self.get_geometry(index)) is None:
-                continue
+        for geometry in self.get_selected_geometries():
             try:
                 geometry._meta["fit"].vertices
                 formats.extend(formats_extended)
@@ -352,6 +378,20 @@ class DataContainerInteractor(QObject):
             representation_menu.addAction(action)
         context_menu.addMenu(representation_menu)
 
+        context_menu.addSeparator()
+
+        group_action = QAction("Group", self.data_list)
+        group_action.triggered.connect(
+            lambda: self.data_list.group_selected("New Group")
+        )
+        context_menu.addAction(group_action)
+
+        ungroup_action = QAction("Ungroup", self.data_list)
+        ungroup_action.triggered.connect(
+            lambda: (self.data_list.ungroup_selected(), self.render())
+        )
+        context_menu.addAction(ungroup_action)
+
         context_menu.addSeparator()
         export_menu = QAction("Export As", self.data_list)
         export_menu.triggered.connect(lambda: self._handle_export())
@@ -366,10 +406,20 @@ class DataContainerInteractor(QObject):
     def _handle_export(self, *args, **kwargs):
         from .dialogs import ExportDialog
 
+        enabled_categories = ["pointcloud", "volume"]
+        for geometry in self.get_selected_geometries():
+            fit = geometry._meta.get("fit", None)
+            if hasattr(fit, "mesh"):
+                enabled_categories.append("mesh")
+
         # Shape and sampling are given if MosaicData.open_file loaded a volume.
         # For convenience, outputs will be handled w.r.t to the initial volume
         sampling, shape = 1, self.container.metadata.get("shape")
-        dialog = ExportDialog(parent=None, parameters={"shape": shape})
+        dialog = ExportDialog(
+            parent=None,
+            parameters={"shape": shape},
+            enabled_categories=enabled_categories,
+        )
 
         if shape is not None:
             sampling = self.container.metadata.get("sampling_rate", 1)
@@ -393,138 +443,24 @@ class DataContainerInteractor(QObject):
             return -1
 
         try:
-            self._export_data(file_path, export_data)
+            _export_data(
+                geometries=self.get_selected_geometries(),
+                metadata=self.container.metadata,
+                file_path=file_path,
+                export_data=export_data,
+            )
         except Exception as e:
             QMessageBox.warning(None, "Error", str(e))
         return None
 
-    def _export_data(self, file_path: str, export_data: Dict):
-        from .utils import points_to_volume
-        from .formats import OrientationsWriter, write_density
-
-        mesh_formats = ("obj", "stl", "ply")
-        volume_formats = ("mrc", "em", "h5")
-        point_formats = ("tsv", "star", "xyz")
-
-        file_format = export_data.get("format")
-        indices = self._get_selected_indices()
-        if not len(indices):
-            raise ValueError("Select at least one object.")
-
-        file_path, _ = splitext(file_path)
-
-        shape = self.container.metadata.get("shape", None)
-        if shape is not None:
-            _sampling = self.container.metadata.get("sampling_rate", 1)
-            shape = np.rint(np.divide(shape, _sampling)).astype(int)
-
-        center = 0
-        if {"shape_x", "shape_y", "shape_z"}.issubset(export_data):
-            shape = tuple(export_data[x] for x in ["shape_x", "shape_y", "shape_z"])
-
-        orientation_kwargs = {}
-        data = {"points": [], "quaternions": []}
-        for index in indices:
-            geometry = self.container.get(index)
-            if geometry is None:
-                continue
-
-            if file_format in mesh_formats:
-                fit = geometry._meta.get("fit", None)
-                if not hasattr(fit, "mesh"):
-                    raise ValueError(f"Selected geometry {index} is not a mesh.")
-                fit.to_file(f"{file_path}_{index}.{file_format}")
-                continue
-
-            points, quaternions = geometry.points, None
-            if file_format in point_formats:
-                points, quaternions = geometry.points, geometry.quaternions
-
-            if quaternions is None:
-                quaternions = np.full((points.shape[0], 4), fill_value=(1, 0, 0, 0))
-
-            sampling = export_data.get("sampling", -1)
-            if sampling < 0:
-                sampling = geometry.sampling_rate
-
-            if export_data.get("relion_5_format", False):
-                center = np.divide(shape, 2).astype(int) if shape is not None else 0
-                center = np.multiply(center, sampling)
-                orientation_kwargs["version"] = "# version 50001"
-                sampling = 1
-
-            points = np.subtract(np.divide(points, sampling), center)
-            data["points"].append(points)
-            data["quaternions"].append(quaternions)
-
-        if file_format in mesh_formats:
-            return None
-
-        if len(data["points"]) == 0:
-            raise ValueError("No elements suitable for export")
-
-        if file_format in volume_formats:
-            if shape is None:
-                temp = np.rint(np.concatenate(data["points"]))
-                shape = temp.astype(int).max(axis=0) + 1
-
-            volume = None
-            for index, points in enumerate(data["points"]):
-                volume = points_to_volume(
-                    points, sampling_rate=1, shape=shape, weight=index + 1, out=volume
-                )
-
-            # Try saving some memory on write. uint8 would be padded to 16 hence int8
-            dtype = np.float32
-            if index < np.iinfo(np.int8).max:
-                dtype = np.int8
-            elif index < np.iinfo(np.uint16).max:
-                dtype = np.uint16
-
-            return write_density(
-                volume.astype(dtype),
-                filename=f"{file_path}.{file_format}",
-                sampling_rate=sampling,
-            )
-
-        if file_format not in point_formats:
-            return None
-
-        data["entities"] = [
-            np.full(x.shape[0], fill_value=i) for i, x in enumerate(data["points"])
-        ]
-        if single_file := export_data.get("single_file", True):
-            data = {k: [np.concatenate(v)] for k, v in data.items()}
-
-        if file_format == "xyz":
-            for index, points in enumerate(data["points"]):
-                fname = f"{file_path}_{index}.{file_format}"
-                if single_file:
-                    fname = f"{file_path}.{file_format}"
-
-                header = ""
-                if export_data.get("header", True):
-                    header = ",".join(["x", "y", "z"])
-
-                np.savetxt(fname, points, delimiter=",", header=header, comments="")
-            return 1
-
-        for index in range(len(data["points"])):
-            orientations = OrientationsWriter(**{k: v[index] for k, v in data.items()})
-            fname = f"{file_path}_{index}.{file_format}"
-            if single_file:
-                fname = f"{file_path}.{file_format}"
-            orientations.to_file(fname, file_format=file_format, **orientation_kwargs)
-
     def _show_properties_dialog(self) -> int:
         from .dialogs import GeometryPropertiesDialog
 
-        indices = self._get_selected_indices()
-        indices = [x for x in indices if self.container._index_ok(x)]
-        if not len(indices):
+        uuids = self._get_selected_uuids()
+        if not len(uuids):
             return -1
 
-        base_container = self.container.data[indices[0]]
+        base_container = self.container.get(uuids[0])
         base_parameters = base_container._appearance.copy()
         base_parameters["sampling_rate"] = base_container.sampling_rate
 
@@ -532,15 +468,17 @@ class DataContainerInteractor(QObject):
 
         def on_parameters_changed(parameters):
             sampling_rate = parameters.pop("sampling_rate")
-            full_render = self.container.update_appearance(indices, parameters)
-            for index in indices:
-                self.container.data[index]._sampling_rate = sampling_rate
+            full_render = self.container.update_appearance(uuids, parameters)
+            for uuid in uuids:
+                if (geometry := self.container.get(uuid)) is None:
+                    continue
+                geometry.sampling_rate = sampling_rate
 
             if full_render:
                 self.render()
 
             # Make sure selection is maintained and invoke render_vtk afterwards
-            self.set_selection(indices)
+            self.set_selection_by_uuid(uuids)
 
         dialog.parametersChanged.connect(on_parameters_changed)
 
@@ -548,25 +486,38 @@ class DataContainerInteractor(QObject):
             on_parameters_changed(base_parameters)
         return 1
 
-    def remove_points(self):
-        added_cluster = self.add_selection(self.point_selection)
-        if added_cluster == -1:
-            return -1
-        self.point_selection.clear()
-        self.container.remove(added_cluster)
+    def _uuid_to_items(self):
+        uuid_to_items = {}
+        for i in range(len(self.container)):
+            if (geometry := self.container.get(i)) is None:
+                continue
 
-    def cluster_points(self):
-        ret = self.add_selection(self.point_selection)
-        self.deselect_points()
-        return ret
+            name = geometry._meta.get("name", None)
+            if name is None:
+                name = f"{self.prefix} {i}"
+
+            item_type = "cluster"
+            if (fit := geometry._meta.get("fit", None)) is not None:
+                item_type = "mesh" if hasattr(fit, "mesh") else "parametric"
+                if item_type == "mesh" and hasattr(geometry, "_trajectory"):
+                    item_type = "trajectory"
+
+            info = {"item_type": item_type, "name": name, "uuid": geometry.uuid}
+
+            geometry._meta["name"] = name
+            geometry._meta["info"] = info
+
+            item = StyledTreeWidgetItem(
+                name, geometry.visible, info, editable=True, parent=None
+            )
+            uuid_to_items[geometry.uuid] = item
+        return uuid_to_items
 
     def render(self):
-        from .widgets import StyledListWidgetItem
-
+        """Synchronize vtk actors and tree data structure with subsequent render."""
         renderer = self.vtk_widget.GetRenderWindow().GetRenderers().GetFirstRenderer()
 
         current_actors = set(self.container.get_actors())
-
         actors_to_remove = self.rendered_actors - current_actors
         for actor in actors_to_remove:
             renderer.RemoveActor(actor)
@@ -577,75 +528,38 @@ class DataContainerInteractor(QObject):
             renderer.AddActor(actor)
             self.rendered_actors.add(actor)
 
-        self.data_list.clear()
-        for i in range(len(self.container)):
-            name = self.container.data[i]._meta.get("name", None)
-            if name is None:
-                name = f"{self.prefix} {i}"
-
-            visible = self.container.data[i].visible
-            text = self.container.data[i]._meta.get("metadata_text", None)
-
-            type = "cluster"
-            geometry = self.container.data[i]
-            fit = geometry._meta.get("fit", None)
-            if fit is not None:
-                text = "mesh" if hasattr(fit, "mesh") else "parametric"
-                if text == "mesh" and hasattr(geometry, "_trajectory"):
-                    text = "trajectory"
-                type = text
-
-            meta_info = geometry._meta.pop("info", {})
-            meta_info.update({"metadata_text": text, "item_type": type, "name": name})
-
-            geometry._meta["info"] = meta_info
-            item = StyledListWidgetItem(name, visible, meta_info, editable=True)
-            self.data_list.addItem(item)
+        uuid_to_items = self._uuid_to_items()
+        self.data_list.update(uuid_to_items)
 
         self.render_vtk()
-        return self.render_update.emit()
+        self.render_update.emit()
 
     def render_vtk(self):
-        # TODO: Render needs an update to not require container and data_list
-        # to be constantly synced. Then this can be moved into render
-        try:
-            self.data_list.list_widget.blockSignals(True)
-            for i in range(self.data_list.count()):
-                if (geometry := self.container.get(i)) is None:
-                    continue
-                item = self.data_list.item(i)
-                item.set_visible(geometry.visible)
-        except Exception as e:
-            print(e)
-            pass
-        finally:
-            self.data_list.list_widget.blockSignals(False)
-
+        """Update the vtk scene."""
         self.vtk_pre_render.emit()
         return self.vtk_widget.GetRenderWindow().Render()
 
     def deselect_points(self):
-        for cluster_index, point_ids in self.point_selection.items():
-            if (geometry := self.get_geometry(cluster_index)) is None:
+        for uuid, point_ids in self.point_selection.items():
+            if (geometry := self.container.get(uuid)) is None:
                 continue
 
             color = geometry._appearance.get("base_color", (0.7, 0.7, 0.7))
-            self.container.highlight_points(cluster_index, point_ids, color)
+            self.container.highlight_points(uuid, point_ids, color)
 
         self.point_selection.clear()
 
     def highlight_selected_points(self, color):
-        for cluster_index, point_ids in self.point_selection.items():
-            self.container.highlight_points(cluster_index, point_ids, color)
-
-        self.vtk_widget.GetRenderWindow().Render()
+        for uuid, point_ids in self.point_selection.items():
+            self.container.highlight_points(uuid, point_ids, color)
+        return self.render_vtk()
 
     def highlight_clusters_from_selected_points(self):
-        return self.set_selection(list(self.point_selection.keys()))
+        return self.set_selection_by_uuid(list(self.point_selection.keys()))
 
     def change_representation(self, representation: str):
-        indices = self._get_selected_indices()
-        if not len(indices):
+        geometries = self.get_selected_geometries()
+        if not len(geometries):
             return -1
 
         representation = representation.lower().replace(" ", "_")
@@ -653,30 +567,27 @@ class DataContainerInteractor(QObject):
         if representation == "points":
             representation = "pointcloud"
 
-        for index in indices:
-            if (geometry := self.get_geometry(index)) is None:
-                continue
-
+        for geometry in geometries:
             # BUG: Moving from pointcloud_normals to a different representation and
             # back breaks glyph rendering. This could be due to incorrect cleanup in
             # Geometry.change_representation or an issue of vtk 9.3.1. Creating a copy
             # of the Geometry instance circumvents the issue.
             if representation in ("normals", "basis", "gaussian_density"):
-                self.container.data[index] = geometry[...]
-                geometry = self.container.data[index]
+                self.container.update(geometry.uuid, geometry[...])
+                geometry = self.container.get(geometry.uuid)
 
             geometry.change_representation(representation)
         self.render()
 
     def _backup(self):
         # Save clusters and points that are modified by the operation
-        self._merge_index = None
+        self._merge_uuid = None
         try:
             self._geometry_backup = {
-                i: self.get_geometry(i)[...] for i in self._get_selected_indices()
+                x.uuid: x[...] for x in self.get_selected_geometries()
             }
             self._point_backup = {
-                i: self.get_geometry(i)[ix] for i, ix in self.point_selection.items()
+                i: self.container.get(i)[ix] for i, ix in self.point_selection.items()
             }
         except Exception:
             self._geometry_backup = None
@@ -686,86 +597,82 @@ class DataContainerInteractor(QObject):
         if getattr(self, "_geometry_backup", None) is None:
             return None
 
-        if getattr(self, "_merge_index", None) is not None:
-            self.container.remove(self._merge_index)
+        if getattr(self, "_merge_uuid", None) is not None:
+            self.container.remove(self._merge_uuid)
 
-        for i, geometry in self._geometry_backup.items():
+        for _, geometry in self._geometry_backup.items():
             self.add(geometry)
 
-        for i, geometry in self._point_backup.items():
-            if not self.container._index_ok(i):
+        for uuid, geometry in self._point_backup.items():
+            prev_geometry = self.container.get(uuid)
+            if prev_geometry is None:
                 self.add(geometry)
                 continue
-            self.container.data[i] = geometry.merge((geometry, self.container.data[i]))
+            self.container.update(uuid, geometry.merge((prev_geometry, geometry)))
 
         self._geometry_backup = None
         self._point_backup = None
-        self._merge_index = None
+        self._merge_uuid = None
         self.data_changed.emit()
         self.render()
 
     def merge(self):
+        from .geometry import Geometry
+
         self._backup()
-        point_cluster = self.cluster_points()
-        new_cluster = self.merge_cluster()
-        self.merge_cluster(indices=(new_cluster, point_cluster))
-        self._merge_index = len(self.container.data) - 1
-        self.data_changed.emit()
+        point_cluster = self.add_selection(self.point_selection)
+        self.deselect_points()
+
+        merge = [*self.get_selected_geometries(), self.container.get(point_cluster)]
+        merged_geometry = Geometry.merge(merge)
+
+        self.container.remove(merge)
+        new_index = self.add(merged_geometry)
+        self._merge_uuid = self.container.get(new_index).uuid
+
+        self.render()
 
     def remove(self):
         self._backup()
-        self.remove_points()
-        self.remove_cluster()
+
+        added_cluster = self.add_selection(self.point_selection)
+        self.point_selection.clear()
+
+        self.container.remove(
+            [self.container.get(added_cluster), *self.get_selected_geometries()]
+        )
         self.data_changed.emit()
+        self.render()
 
     def refresh_actors(self):
         for index in range(len(self.container)):
-            self.container.data[index] = self.container.data[index][...]
+            geometry = self.container.get(index)
+            self.container.update(geometry, geometry[...])
         return self.render()
 
-    def remove_cluster(self, **kwargs):
-        self.container.remove(self._get_selected_indices())
-        return self.render()
+    def update(self, container, tree_state=None):
+        """Update container with new data and optionally restore tree structure.
 
-    def merge_cluster(self, **kwargs):
-        from .geometry import Geometry
-
-        indices = self._get_selected_indices()
-        kwarg_indices = kwargs.pop("indices", ())
-        if not isinstance(kwarg_indices, (Tuple, List)):
-            kwarg_indices = [kwarg_indices]
-
-        indices = (*indices, *kwarg_indices)
-        geometries = [self.container.get(i) for i in (*indices, *kwarg_indices)]
-        geometries = [x for x in geometries if x is not None]
-        if len(geometries) == 0:
-            return None
-
-        geometry = Geometry.merge(geometries)
-
-        self.container.remove(indices)
-        self.add(geometry)
-        self.render()
-
-    def update(self, other):
-        if not isinstance(other, type(self.container)):
+        Parameters
+        ----------
+        container : :py:class:`mosaic.container.DataContainer`
+            Container with new data
+        tree_state : TreeState, optional
+            Tree structure to restore. If None, items added to root.
+        """
+        if not isinstance(container, type(self.container)):
             raise ValueError(
-                f"Can not update {type(self.container)} using {type(other)}."
+                f"Can not update {type(self.container)} using {type(container)}."
             )
 
         self.container.clear()
-        self.container.metadata.update(other.metadata)
-        for index in range(len(other)):
-            geometry = other.get(index)
-            if geometry is None:
-                continue
-            self.add(geometry)
-        self.data_changed.emit()
+        self.container.metadata.update(container.metadata)
+        _ = [self.add(x) for x in container.data]
 
-    def deselect(self):
-        """Deselect on right-click"""
-        self.data_list.clearSelection()
-        self.deselect_points()
+        if tree_state is not None:
+            self.data_list.apply_state(tree_state, self._uuid_to_items())
+
+        self.data_changed.emit()
 
 
 _GEOMETRY_OPERATIONS = {
@@ -777,7 +684,7 @@ _GEOMETRY_OPERATIONS = {
     "duplicate": {"remove_original": False},
     "visibility": {
         "remove_original": False,
-        "render": "vtk",
+        "render": "full",
         "background": False,
         "batch": True,
     },
@@ -795,21 +702,20 @@ for op_name, config in _GEOMETRY_OPERATIONS.items():
             f"""Apply {op_name} operation to selected geometries in background."""
             from .operations import GeometryOperations
 
-            selected_indices = kwargs.get("geometry_indices")
-            if selected_indices is None:
-                selected_indices = self._get_selected_indices()
-
             def _render_callback(*args, **kwargs):
                 if render_flag == "full":
                     self.render()
                 elif render_flag == "vtk":
                     self.render_vtk()
 
-            for index in selected_indices:
-                if (geometry := self.get_geometry(index)) is None:
+            if (geometries := kwargs.get("geometries", None)) is None:
+                geometries = self.get_selected_geometries()
+
+            for geometry in geometries:
+                if geometry is None:
                     continue
 
-                def _callback(ret):
+                def _callback(ret, geom=geometry):
                     if ret is None:
                         pass
                     elif isinstance(ret, (List, Tuple)):
@@ -818,7 +724,7 @@ for op_name, config in _GEOMETRY_OPERATIONS.items():
                         self.add(ret)
 
                     if remove_orig:
-                        self.container.remove(geometry)
+                        self.container.remove(geom)
 
                     if not batch_flag:
                         _render_callback()
@@ -842,3 +748,117 @@ for op_name, config in _GEOMETRY_OPERATIONS.items():
         method_name,
         create_method(op_name, remove_orig, render, background, batch),
     )
+
+
+def _export_data(geometries, metadata, file_path: str, export_data: Dict):
+    from .utils import points_to_volume
+    from .formats import OrientationsWriter, write_density
+
+    mesh_formats = ("obj", "stl", "ply")
+    volume_formats = ("mrc", "em", "h5")
+    point_formats = ("tsv", "star", "xyz")
+
+    file_format = export_data.get("format")
+    if not len(geometries):
+        raise ValueError("Select at least one object.")
+
+    file_path, _ = splitext(file_path)
+
+    shape = metadata.get("shape", None)
+    if shape is not None:
+        _sampling = metadata.get("sampling_rate", 1)
+        shape = np.rint(np.divide(shape, _sampling)).astype(int)
+
+    center = 0
+    if {"shape_x", "shape_y", "shape_z"}.issubset(export_data):
+        shape = tuple(export_data[x] for x in ["shape_x", "shape_y", "shape_z"])
+
+    orientation_kwargs = {}
+    data = {"points": [], "quaternions": []}
+    for index, geometry in enumerate(geometries):
+        if file_format in mesh_formats:
+            fit = geometry._meta.get("fit", None)
+            if not hasattr(fit, "mesh"):
+                raise ValueError(f"Selected geometry {index} is not a mesh.")
+            fit.to_file(f"{file_path}_{index}.{file_format}")
+            continue
+
+        points, quaternions = geometry.points, None
+        if file_format in point_formats:
+            points, quaternions = geometry.points, geometry.quaternions
+
+        if quaternions is None:
+            quaternions = np.full((points.shape[0], 4), fill_value=(1, 0, 0, 0))
+
+        sampling = export_data.get("sampling", -1)
+        if sampling < 0:
+            sampling = geometry.sampling_rate
+
+        if export_data.get("relion_5_format", False):
+            center = np.divide(shape, 2).astype(int) if shape is not None else 0
+            center = np.multiply(center, sampling)
+            orientation_kwargs["version"] = "# version 50001"
+            sampling = 1
+
+        points = np.subtract(np.divide(points, sampling), center)
+        data["points"].append(points)
+        data["quaternions"].append(quaternions)
+
+    if file_format in mesh_formats:
+        return None
+
+    if len(data["points"]) == 0:
+        raise ValueError("No elements suitable for export")
+
+    if file_format in volume_formats:
+        if shape is None:
+            temp = np.rint(np.concatenate(data["points"]))
+            shape = temp.astype(int).max(axis=0) + 1
+
+        volume = None
+        for index, points in enumerate(data["points"]):
+            volume = points_to_volume(
+                points, sampling_rate=1, shape=shape, weight=index + 1, out=volume
+            )
+
+        # Try saving some memory on write. uint8 would be padded to 16 hence int8
+        dtype = np.float32
+        if index < np.iinfo(np.int8).max:
+            dtype = np.int8
+        elif index < np.iinfo(np.uint16).max:
+            dtype = np.uint16
+
+        return write_density(
+            volume.astype(dtype),
+            filename=f"{file_path}.{file_format}",
+            sampling_rate=sampling,
+        )
+
+    if file_format not in point_formats:
+        return None
+
+    data["entities"] = [
+        np.full(x.shape[0], fill_value=i) for i, x in enumerate(data["points"])
+    ]
+    if single_file := export_data.get("single_file", True):
+        data = {k: [np.concatenate(v)] for k, v in data.items()}
+
+    if file_format == "xyz":
+        for index, points in enumerate(data["points"]):
+            fname = f"{file_path}_{index}.{file_format}"
+            if single_file:
+                fname = f"{file_path}.{file_format}"
+
+            header = ""
+            if export_data.get("header", True):
+                header = ",".join(["x", "y", "z"])
+
+            np.savetxt(fname, points, delimiter=",", header=header, comments="")
+        return 1
+
+    for index in range(len(data["points"])):
+        orientations = OrientationsWriter(**{k: v[index] for k, v in data.items()})
+        fname = f"{file_path}_{index}.{file_format}"
+        if single_file:
+            fname = f"{file_path}.{file_format}"
+        orientations.to_file(fname, file_format=file_format, **orientation_kwargs)
