@@ -17,6 +17,7 @@ from qtpy.QtWidgets import (
     QPushButton,
     QFileDialog,
     QLabel,
+    QGroupBox,
 )
 import qtawesome as qta
 
@@ -51,8 +52,8 @@ class VolumeViewer(QWidget):
         self.slice = vtk.vtkImageSlice()
         self.volume = None
 
-        self.label = QLabel("Volume")
-        self.open_button = QPushButton("Open")
+        # self.label = QLabel("Volume Viewer")
+        self.open_button = QPushButton("Load")
         self.open_button.clicked.connect(self.open_volume)
         self.close_button = QPushButton("Close")
         self.close_button.clicked.connect(self.close_volume)
@@ -119,7 +120,7 @@ class VolumeViewer(QWidget):
 
         # Create layout
         self.controls_layout = QHBoxLayout()
-        self.controls_layout.addWidget(self.label)
+        # self.controls_layout.addWidget(self.label)
         self.controls_layout.addWidget(self.open_button)
         self.controls_layout.addWidget(self.close_button)
         self.controls_layout.addWidget(self.orientation_selector)
@@ -236,21 +237,11 @@ class VolumeViewer(QWidget):
         from ..formats.parser import load_density
 
         volume = load_density(file_path)
-
         self.volume = vtk.vtkImageData()
         self.volume.SetDimensions(volume.shape)
         self.volume.SetSpacing(volume.sampling_rate)
         self.volume.AllocateScalars(vtk.VTK_FLOAT, 1)
 
-        # if volume.shape[1] == 20:
-        #     print("Change this back")
-
-        #     self.volume.SetOrigin(
-        #         -2 * 4.0,
-        #         (1320 -2) * volume.sampling_rate[1],
-        #         -1 * 5.24
-        #     )
-
         volume = numpy_support.numpy_to_vtk(
             volume.data.ravel(order="F"), deep=True, array_type=vtk.VTK_FLOAT
         )
@@ -305,6 +296,12 @@ class VolumeViewer(QWidget):
         self.renderer.ResetCamera()
         self.vtk_widget.GetRenderWindow().Render()
 
+    def get_slice(self):
+        try:
+            return int(self.slice_value_label.text())
+        except Exception:
+            return 0
+
     def get_orientation(self):
         return getattr(self, "_orientation", None)
 
@@ -366,25 +363,38 @@ class VolumeViewer(QWidget):
             *[0 if i != dim else origin + pos * spacing for i in range(3)]
         )
 
-        return None
+    def remove_existing_clipping_plane(self, mapper):
+        if (planes := mapper.GetClippingPlanes()) is None:
+            return None
 
-    def handle_projection_change(self, state):
+        planes.InitTraversal()
+        for j in range(planes.GetNumberOfItems()):
+            plane = planes.GetNextItem()
+            if plane == self.clipping_plane:
+                mapper.RemoveClippingPlane(self.clipping_plane)
+
+    def handle_projection_change(self, state=None):
         if self.volume is None:
             return
 
+        if state is None:
+            state = self.project_selector.currentText()
+
         self._projection = state
         actors = self.renderer.GetActors()
         actors.InitTraversal()
 
         for i in range(actors.GetNumberOfItems()):
             actor = actors.GetNextActor()
-            actor.GetMapper().RemoveAllClippingPlanes()
+            mapper = actor.GetMapper()
+
+            self.remove_existing_clipping_plane(mapper)
             if state == "Off":
                 continue
 
             self.clipping_direction = 1 if state == "Project +" else -1
             self.update_clipping_plane()
-            actor.GetMapper().AddClippingPlane(self.clipping_plane)
+            mapper.AddClippingPlane(self.clipping_plane)
 
         self.vtk_widget.GetRenderWindow().Render()
 
@@ -407,8 +417,14 @@ class MultiVolumeViewer(QWidget):
         )
 
         self.layout = QVBoxLayout(self)
-        self.layout.setSpacing(5)
-        self.layout.setContentsMargins(0, 0, 0, 5)
+        self.layout.setSpacing(0)
+        self.layout.setContentsMargins(4, 4, 4, 4)
+
+        self.viewer_group = QGroupBox("Volume Viewer")
+        self.viewer_layout = QVBoxLayout(self.viewer_group)
+        self.layout.addWidget(self.viewer_group)
+        self.viewer_layout.setSpacing(4)
+        self.viewer_layout.setContentsMargins(0, 4, 0, 4)
 
         self.primary = VolumeViewer(self.vtk_widget, self.legend)
         current_margins = self.primary.layout().contentsMargins()
@@ -416,16 +432,17 @@ class MultiVolumeViewer(QWidget):
             current_margins.left(), 0, current_margins.right(), 0
         )
         self.primary_margins = self.primary.layout().contentsMargins()
-        self.layout.addWidget(self.primary)
+        self.viewer_layout.addWidget(self.primary)
+        # self.layout.addWidget(self.primary)
 
         add_button = QPushButton()
         add_button.setIcon(qta.icon("fa5s.plus", color="#696c6f"))
         add_button.setFixedWidth(30)
         add_button.clicked.connect(self.add_viewer)
         self.primary.controls_layout.addWidget(add_button)
+        self.primary.close_button.clicked.connect(self._promote_new_primary)
 
         self.additional_viewers = []
-        self.primary.data_changed.connect(self._changed_primary)
 
     def add_viewer(self):
         """Add a new VolumeViewer instance"""
@@ -446,7 +463,7 @@ class MultiVolumeViewer(QWidget):
 
         self._copy_from_primary(new_viewer)
         self.additional_viewers.append(new_viewer)
-        self.layout.addWidget(new_viewer)
+        self.viewer_layout.addWidget(new_viewer)
 
     def remove_viewer(self, viewer):
         """Remove a specific viewer"""
@@ -458,6 +475,10 @@ class MultiVolumeViewer(QWidget):
     def close(self):
         for viewer in self.additional_viewers:
             viewer.close_volume()
+        try:
+            self.primary.close_button.clicked.disconnect(self._promote_new_primary)
+        except TypeError:
+            pass
         self.primary.close_volume()
 
     def _copy_from_primary(self, new_viewer: VolumeViewer) -> int:
@@ -468,5 +489,35 @@ class MultiVolumeViewer(QWidget):
 
         return new_viewer.swap_volume(volume)
 
-    def _changed_primary(self):
-        return [self._copy_from_primary(x) for x in self.additional_viewers]
+    def _promote_new_primary(self) -> int:
+        viewers = [
+            x for x in self.additional_viewers if getattr(x, "volume") is not None
+        ]
+
+        if not len(viewers):
+            return None
+
+        new_primary = viewers[0]
+
+        # Copy all state from the viewer being promoted
+        self.primary.swap_volume(new_primary.volume)
+        self.primary.change_orientation(new_primary.get_orientation())
+        self.primary.update_slice(new_primary.get_slice())
+        self.primary.handle_projection_change(new_primary.get_projection())
+
+        # Copy visual settings
+        self.primary.color_selector.setCurrentText(
+            new_primary.color_selector.currentText()
+        )
+        self.primary.min_contrast_slider.setValue(
+            new_primary.min_contrast_slider.value()
+        )
+        self.primary.max_contrast_slider.setValue(
+            new_primary.max_contrast_slider.value()
+        )
+        self.primary.gamma_slider.setValue(new_primary.gamma_slider.value())
+
+        if new_primary.is_visible != self.primary.is_visible:
+            self.primary.toggle_visibility()
+
+        self.remove_viewer(new_primary)
